#!/usr/bin/env python
#
# results_load.py - The fsleyes MRS plugin - load results from fsl_mrsi
#
# Author: Will Clarke <william.clarke@ndcn.ox.ac.uk>
#

import os
from pathlib import Path
import logging

import numpy as np
import wx

import fsl.utils.settings as fslsettings
import fsl.data.image as fslimage

import fsleyes.actions as actions
import fsleyes_widgets.utils.status as status
from fsleyes.views.orthopanel import OrthoPanel
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes_props as props

from fsleyes_plugin_mrs.plugin import MRSView

log = logging.getLogger(__name__)


###########################################
# FSLFitTool - load results from fsl_mrsi #
###########################################
class FSLFitTool(actions.Action):
    """Load results as output from fsl_mrsi.
    Load and display fit in the MRSView display
    Add a FSLMRSResultsControl panel to the ortho view to allow
    easy loading of metabolite maps
    """
    def __init__(self, overlayList, displayCtx, frame):
        actions.Action.__init__(self, overlayList, displayCtx, self.__loadResults)
        self.frame = frame

    def __loadResults(self):
        """Show a dialog prompting the user for a directory to load.
        Subsequently interprets results structure, causes a FSLMRSResultsControl
        to be created and then load the fit, baseline, and residual overlays.
        """

        msg = 'Load fsl_mrs(i) fit directory'
        fromDir = fslsettings.read('loadSaveOverlayDir', os.getcwd())
        dlg = wx.DirDialog(wx.GetApp().GetTopWindow(),
                           message=msg,
                           defaultPath=fromDir,
                           style=wx.FD_OPEN)

        if dlg.ShowModal() != wx.ID_OK:
            return

        dirPath = Path(dlg.GetPath())
        errtitle = 'Error loading directory'
        errmsg = 'An error occurred while loading the fit directory.'
        with status.reportIfError(errtitle, errmsg, raiseError=False):
            # Steps:
            # 1. Look through the directory and identify metabolite names
            # and things that can be loaded
            self._identifyResults(dirPath)
            # 2. Create a panel to control the loading of the possible views
            self._createResultsPanel()
            # 3. Setup the ortho panel ready for the display
            self._set_up_ortho()
            # 4. Load the fit, baseline and residual into the MRS view.
            self._loadFit(dirPath)

    def _identifyResults(self, dir_path):
        """Identify the metabolites used and the images availible.
        Store this information for use constructing the additional panel

        :param Path dir_path: Path selected by user in DirDialog window
        """
        # print(f'Interpreting fit path {dir_path}')
        # List of metabolites
        self.metab_list = []
        for file in (dir_path / 'concs' / 'raw').glob('*.nii.gz'):
            while file.suffix in {'.nii', '.gz'}:
                file = file.with_suffix('')
            self.metab_list.append(file.name)

        # List of things that could be displayed
        self.display_options = {'conc-raw': dir_path / 'concs/raw/{}.nii.gz',
                                'conc-internal': dir_path / 'concs/internal/{}.nii.gz',
                                'conc-molarity': dir_path / 'concs/molarity/{}.nii.gz',
                                'conc-molality': dir_path / 'concs/molality/{}.nii.gz',
                                'uncertainties': dir_path / 'uncertainties/{}_sd.nii.gz',
                                'SNR': dir_path / 'qc/{}_snr.nii.gz',
                                'FWHM': dir_path / 'qc/{}_fwhm.nii.gz'}

    def _createResultsPanel(self):
        """Activate a FSLMRSResultsControl panel, update with the metablist and display options."""
        # Activate the panel
        ortho = self.frame.getView(OrthoPanel)
        if len(ortho) == 0:
            log.error('No Ortho panel present')
            return
        else:
            # Assume only 1 ortho panel
            ortho = ortho[0]

        ortho.togglePanel(FSLMRSResultsControl)

        # Update the properties
        ortho.getPanel(FSLMRSResultsControl).update_choices(self.metab_list, self.display_options)

    def _set_up_ortho(self):
        """Add colour bar to ortho panel."""
        ortho = self.frame.getView(OrthoPanel)
        if len(ortho) == 0:
            log.error('No Ortho panel present')
            return
        else:
            # Assume only 1 ortho panel
            ortho = ortho[0]

        orthoOpts = ortho.sceneOpts
        orthoOpts.showColourBar = True
        orthoOpts.colourBarLocation = 'left'
        orthoOpts.colourBarLabelSide = 'top-left'
        orthoOpts.labelSize = 10

    def _loadFit(self, dir_path):
        """Load the fit, baseline, and residual data from the fit subdir.
        Add to the overlayList and display in the MRSView panel.
        """
        def load_and_disp(file_name, lw, colour):
            # Load and add to the overlayList
            full_path = dir_path / 'fit' / f'{file_name}.nii.gz'
            if not full_path.exists():
                log.warn(f'{file_name} image not found in {full_path}, skipping.')
                return
            new_img = fslimage.Image(str(full_path))
            self.overlayList.append(new_img)

            # Loop through the viewPanels. If MRSView enable overlay, otherwise disable overlay
            for panel in self.frame.viewPanels:
                display = panel.displayCtx.getDisplay(self.overlayList[-1])
                if isinstance(panel, MRSView):
                    display.enabled = True

                    # Set colour etc.
                    # Get the data series
                    ps = panel.getDataSeries(self.overlayList[-1])
                    ps.lineStyle = '-'
                    ps.alpha = 1.0
                    ps.lineWidth = lw
                    ps.colour = colour
                else:
                    display.enabled = False

        # Load the fit scan
        load_and_disp('fit', 2, (1, 0, 0))
        # Load the baseline scan
        load_and_disp('baseline', 1, (0, 0.5, 0.5))
        # Load the fit scan
        load_and_disp('residual', 1, (0.2, 0.2, 0.2))


class FSLMRSResultsControl(ctrlpanel.SettingsPanel):
    """Control panel for the FSL-MRS results. Allows user to easily select overlays to plot
    """

    @staticmethod
    def title():
        """Overrides :meth:`.ControlMixin.title`. Returns a title to be used
        in FSLeyes menus.
        """
        return 'FSL-MRS Results'

    @staticmethod
    def defaultLayout():
        """Overrides :meth:`.ControlMixin.defaultLayout`. Returns arguments
        to be passed to :meth:`.ViewPanel.defaultLayout`.
        """
        return {'location': wx.BOTTOM}

    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``FSLMRSResultsControl`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        return [OrthoPanel]

    def __init__(self, parent, overlayList, displayCtx, viewPanel):
        """Create a ``FSLMRSResultsControl``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg viewPanel:   The :class:`.ViewPanel` instance.
        :arg metabolites:   A list of possible metabolites.
        :arg overlay_options:   A dict of possible overlay types to load.
        """
        super().__init__(parent, overlayList, displayCtx, viewPanel)

        class propStore(props.HasProperties):
            """Class to store properties"""
            metabolite = props.Choice()
            overlay_type = props.Choice()
            replace = props.Boolean(default=True)

        self._propStore = propStore()
        self._overlay_types = None
        self._previous_overlay = None

        self.refreshWidgets()

    def destroy(self):
        """Must be called when this ``FSLMRSResultsControl`` is no
        longer needed. calls the
        :meth:`.SettingsPanel.destroy` method.
        """
        ctrlpanel.SettingsPanel.destroy(self)

    def _generateWidgets(self, group_name):
        '''Make the widgets required for the selection of overlays.'''

        widgetList = self.getWidgetList()

        widgets = []
        metab_sel = props.makeWidget(
                widgetList,
                self._propStore,
                'metabolite')
        widgetList.AddWidget(
                metab_sel,
                displayName='Metabolite',
                tooltip="Select metabolite",
                groupName=group_name)
        widgets.append(metab_sel)

        type_sel = props.makeWidget(
                widgetList,
                self._propStore,
                'overlay_type')
        widgetList.AddWidget(
                type_sel,
                displayName='Type',
                tooltip="Select plot type",
                groupName=group_name)
        widgets.append(type_sel)

        rep_select = props.makeWidget(
                widgetList,
                self._propStore,
                'replace')
        widgetList.AddWidget(
                rep_select,
                displayName='Replace?',
                tooltip="If selected new overlays replace old ones.",
                groupName=group_name)
        widgets.append(rep_select)

        self.__widgets = widgets

    def refreshWidgets(self):
        '''Refresh the widgets for the selection of overlays.
        Run after updating the choices (using update_choices).
        '''
        widgetList = self.getWidgetList()

        if widgetList.HasGroup('fsl_mrs_results'):
            widgetList.RemoveGroup('fsl_mrs_results')

        # Add listeners to the properties which will cause a
        # refresh of the Info Pannel.
        prop_metab = self._propStore.getProp('metabolite')
        prop_metab.addListener(
            self._propStore,
            'metabolite_choice_update',
            self._selected_result_change,
            overwrite=True)

        prop_ot = self._propStore.getProp('overlay_type')
        prop_ot.addListener(
            self._propStore,
            'metabolite_choice_update',
            self._selected_result_change,
            overwrite=True)

        widgetList.AddGroup(
            'fsl_mrs_results',
            'FSL-MRS Results ')

        self._generateWidgets('fsl_mrs_results')

    def update_choices(self, metabolites, overlay_types):
        '''Update the properties (metabolites and overlay types) in the panel.'''
        # Update the metabolite list
        metabolite = self._propStore.getProp('metabolite')
        metabolite.setChoices(metabolites, instance=self._propStore)
        # Update the overlay_types
        overlay_type = self._propStore.getProp('overlay_type')
        overlay_type.setChoices(list(overlay_types.keys()), instance=self._propStore)

        # Store the dictionary defining the types and formats of files.
        self._overlay_types = overlay_types

        # Refresh the widgets to update the GUI
        self.refreshWidgets()

    def _selected_result_change(self, *a):
        """Method called by listeners to load the selected overlay.
        If replace is true / selected then remove the previously loaded
        overlay.
        """
        if self._overlay_types is None:
            return

        full_path = str(self._overlay_types[self._propStore.overlay_type])\
            .format(self._propStore.metabolite)

        if (self._propStore.overlay_type == 'SNR' or self._propStore.overlay_type == 'FWHM')\
                and '+' in self._propStore.metabolite:
            log.info('SNR and FWHM are not defined for combined metabolites.')
            return

        if not Path(full_path).exists():
            log.warn(f'{full_path} image not found, skipping.')
            return

        # If the user has the replace checkbox selected remove any previously loaded
        # overlay
        if self._propStore.replace and self._previous_overlay is not None:
            # Remove previous overlay
            self.overlayList.remove(self._previous_overlay)

        # Load the new overlay
        new_img = fslimage.Image(str(full_path))
        self.overlayList.append(new_img)
        self._previous_overlay = new_img
        self._set_overlay_display(self._propStore.overlay_type)
        self.displayCtx.selectOverlay(new_img)

    def _set_overlay_display(self, type):
        '''Set the display options for the loaded overlay dependent on type.'''

        overlay = self.overlayList[-1]
        display = self.displayCtx.getDisplay(overlay)
        opts = self.displayCtx.getOpts(overlay)

        if 'conc-' in type:
            max_val = np.median(overlay.data[np.nonzero(overlay.data)])\
                 + 3 * np.std(overlay.data[np.nonzero(overlay.data)])
            opts.displayRange = [0.0, max_val]
            opts.cmap = 'hot'
            display.alpha = 67.0
        elif type == 'uncertainties':
            min_val = np.median(overlay.data[np.nonzero(overlay.data)])\
                - 2 * np.std(overlay.data[np.nonzero(overlay.data)])
            max_val = np.median(overlay.data[np.nonzero(overlay.data)])\
                + 2 * np.std(overlay.data[np.nonzero(overlay.data)])
            min_val = np.maximum(min_val, 0.0)
            max_val_clip = np.max(overlay.data[np.nonzero(overlay.data)])
            opts.linkLowRanges = False
            opts.displayRange = [min_val, max_val]
            opts.clippingRange = [0.0, max_val_clip]
            opts.cmap = 'cool'
            display.alpha = 67.0
        elif type == 'SNR':
            min_val = np.median(overlay.data[np.nonzero(overlay.data)])\
                - 2 * np.std(overlay.data[np.nonzero(overlay.data)])
            max_val = np.median(overlay.data[np.nonzero(overlay.data)])\
                + 2 * np.std(overlay.data[np.nonzero(overlay.data)])
            min_val = np.maximum(min_val, 0.0)
            max_val_clip = np.max(overlay.data[np.nonzero(overlay.data)])
            opts.linkLowRanges = False
            opts.displayRange = [min_val, max_val]
            opts.clippingRange = [0.0, max_val_clip]
            opts.cmap = 'red'
            display.alpha = 67.0
        elif type == 'FWHM':
            min_val = np.median(overlay.data[np.nonzero(overlay.data)])\
                - 2 * np.std(overlay.data[np.nonzero(overlay.data)])
            max_val = np.median(overlay.data[np.nonzero(overlay.data)])\
                + 2 * np.std(overlay.data[np.nonzero(overlay.data)])
            min_val = np.maximum(min_val, 0.0)
            max_val_clip = np.max(overlay.data[np.nonzero(overlay.data)])
            opts.linkLowRanges = False
            opts.displayRange = [min_val, max_val]
            opts.clippingRange = [0.0, max_val_clip]
            opts.cmap = 'blue'
            display.alpha = 67.0



#!/usr/bin/env python
#
# plugin.py - The fsleyes MRS plugin.
#
# Author: Will Clarke <william.clarke@ndcn.ox.ac.uk>
#

import logging
import json
import os.path as op
import re

import numpy as np
from numpy import fft
import wx

from fsleyes import __version__ as fsleyes_version
from fsleyes.views.powerspectrumpanel import PowerSpectrumPanel
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.actions as actions
import fsleyes.controls.powerspectrumcontrolpanel as pscontrol
import fsleyes.profiles.shortcuts as fsleyes_shortcuts
import fsleyes_props as props
import fsl.data.image as fslimage
import fsleyes.plotting.powerspectrumseries as psseries
import fsl.utils.cache as cache
from fsleyes.layouts import BUILT_IN_LAYOUTS

# Imports for MRSToolBar
from fsleyes.controls import plottoolbar
import fsleyes.icons as icons
import fsleyes.tooltips as tooltips

from fsleyes.views.orthopanel import OrthoPanel

from fsleyes_plugin_mrs.mrsviewprofile import MRSViewProfile
from fsleyes_plugin_mrs import constants

icon_dir = op.join(op.dirname(__file__), 'icons')
log = logging.getLogger(__name__)


class MRSToolBar(plottoolbar.PlotToolBar):
    """The ``MRSToolBar`` is a toolbar for use with a
    :class:`.MRSView`. It extends :class:`.PlotToolBar`
    mostly replicates :class:`.PowerSpectrumToolBar`
    and adds a few controls specific to the :class:`.PoweSpectrumPanel`.
    """

    @staticmethod
    def title():
        """Overrides :meth:`.ControlMixin.title`. Returns a title to be used
        in FSLeyes menus.
        """
        return 'MRS toolbar'

    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``MRSToolBar`` is only intended to be added to
        :class:`.MRSView` views.
        """
        return [MRSView]

    def __init__(self, parent, overlayList, displayCtx, psPanel):
        """Create a ``MRSToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg psPanel:     The :class:`.PowerSpectrumPanel` instance.
        """

        plottoolbar.PlotToolBar.__init__(
            self, parent, overlayList, displayCtx, psPanel)

        self.togControl = actions.ToggleControlPanelAction(
            overlayList, displayCtx, psPanel, MRSControlPanel)
        self.togDimControl = actions.ToggleControlPanelAction(
            overlayList, displayCtx, psPanel, MRSDimControl)

        togControl = actions.ToggleActionButton(
            'togControl',
            actionKwargs={'floatPane': True},
            icon=[icons.findImageFile('spannerHighlight24'),
                  icons.findImageFile('spanner24')],
            tooltip='Show/hide the MRS control panel.')

        togDimControl = actions.ToggleActionButton(
            'togDimControl',
            icon=[op.join(icon_dir, 'nifti_mrs_icon-mrs_icon_highlight_thumb24.png'),
                  op.join(icon_dir, 'nifti_mrs_icon-mrs_icon_thumb24.png')],
            tooltip='Show/hide the NIfTI-MRS dimension control panel.')

        togList = actions.ToggleActionButton(
            'PlotListPanel',
            actionKwargs={'floatPane': True},
            icon=[icons.findImageFile('listHighlight24'),
                  icons.findImageFile('list24')],
            tooltip=tooltips.actions[psPanel, 'PlotListPanel'])

        togControl = props.buildGUI(self, self, togControl)
        togDimControl = props.buildGUI(self, self, togDimControl)
        togList = props.buildGUI(self, psPanel, togList)

        self.InsertTools([togControl, togDimControl, togList], 0)

        nav = [togControl, togDimControl, togList] + self.getCommonNavOrder()

        self.setNavOrder(nav)


class MRSControlPanel(pscontrol.PowerSpectrumControlPanel):
    """Control panel for the MRS view. Identical to the
    PowerSpectrumControlPanel, but defined so it can be given a distinct name.
    """

    @staticmethod
    def title():
        """Overrides :meth:`.ControlMixin.title`. Returns a title to be used
        in FSLeyes menus.
        """
        return 'MRS control panel'

    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``MRSControlPanel`` is only intended to be added to
        :class:`.MRSView` views.
        """
        return [MRSView]


class MRSDimControl(ctrlpanel.SettingsPanel):
    """Control panel for the MRS view. Controls access to higher dimensions
    of a NIfTI-MRS image.
    """

    @staticmethod
    def title():
        """Overrides :meth:`.ControlMixin.title`. Returns a title to be used
        in FSLeyes menus.
        """
        return 'MRS Dimension control'

    @staticmethod
    def defaultLayout():
        """Overrides :meth:`.ControlMixin.defaultLayout`. Returns arguments
        to be passed to :meth:`.ViewPanel.defaultLayout`.
        """
        return {'location': wx.RIGHT}

    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``MRSDimControl`` is only intended to be added to
        :class:`.MRSView` views.
        """
        return [MRSView]

    def __init__(self, parent, overlayList, displayCtx, plotPanel):
        """Create a ``MRSDimControl``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg psPanel:     The :class:`.PowerSpectrumPanel` instance.
        """
        super().__init__(parent, overlayList, displayCtx, plotPanel)

        self.__plotPanel = plotPanel

        displayCtx.addListener('selectedOverlay',
                               self.name,
                               self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)

        self.__selectedOverlay = None
        self.__selectedOverlayChanged()

    def destroy(self):
        """Must be called when this ``MRSDimControl`` is no
        longer needed. calls the
        :meth:`.PlotControlPanel.destroy` method.
        """
        self.displayCtx.removeListener('selectedOverlay', self.name)
        self.overlayList.removeListener('overlays', self.name)
        ctrlpanel.SettingsPanel.destroy(self)

    def generateDataSeriesWidgets(self, ds, groupName, dims):
        '''Create the required higher dimension spinner widgets in
        the dimension control panel.'''

        widgetList = self.getWidgetList()

        widgets_out = []
        for idx in range(5, 8):
            if dims >= idx:
                dim = props.makeWidget(
                    widgetList,
                    ds,
                    f'dim_{idx}',
                    slider=True,
                    showLimits=False)
                widgetList.AddWidget(
                    dim,
                    displayName=f'DIM {idx}',
                    tooltip=f"DIM {idx} index",
                    groupName=groupName)
                widgets_out.append(dim)

                dim_avg = props.makeWidget(
                    widgetList,
                    ds,
                    f'dim_{idx}_avg')
                widgetList.AddWidget(
                    dim_avg,
                    displayName=f'Average DIM {idx}',
                    tooltip=f"Show average of {idx}th dimension",
                    groupName=groupName)
                widgets_out.append(dim_avg)

                dim_diff = props.makeWidget(
                    widgetList,
                    ds,
                    f'dim_{idx}_diff')
                widgetList.AddWidget(
                    dim_diff,
                    displayName=f'Difference DIM {idx}',
                    tooltip=f"Show difference of {idx}th dimension",
                    groupName=groupName)
                widgets_out.append(dim_diff)

        return widgets_out

    def _set_dim_slider_limits(self):
        """Set the appropriate limits for each dimension index slider.
        If the dim_N_avg property is True then limit to zero."""
        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            return

        ds = self.__plotPanel.getDataSeries(overlay)

        if ds is None:
            return

        for dim in range(5, 8):
            prop = ds.getProp(f'dim_{dim}')
            is_avg = getattr(ds, f'dim_{dim}_avg')
            is_diff = getattr(ds, f'dim_{dim}_diff')

            prop.setAttribute(ds, 'minval', 0)
            if is_avg or is_diff:
                prop.setAttribute(ds, 'maxval', 0)
            elif dim <= overlay.ndim:
                prop.setAttribute(ds, 'maxval', overlay.shape[dim - 1] - 1)
            else:
                prop.setAttribute(ds, 'maxval', 0)

    def _set_dim_diff_enabled(self):
        """Controls whether the difference  checkboxes are enabled or not."""
        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            return

        ds = self.__plotPanel.getDataSeries(overlay)

        if ds is None:
            return

        for dim in range(5, 8):
            prop = ds.getProp(f'dim_{dim}_diff')
            if dim <= overlay.ndim\
                    and overlay.shape[dim - 1] == 2:
                prop.enable(ds)
            else:
                prop.disable(ds)

    def refreshDataSeriesWidgets(self):
        '''Enable/disable and set bounds on the dimensions spinners
        appropriate for the shape of the currently selected overlay.
        '''
        widgetList = self.getWidgetList()

        if self.__selectedOverlay is not None:
            self.__selectedOverlay = None

        if widgetList.HasGroup('niftiMRSDimensions'):
            widgetList.RemoveGroup('niftiMRSDimensions')

        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            return

        ds = self.__plotPanel.getDataSeries(overlay)

        if ds is None:
            return

        self.__selectedOverlay = overlay

        # Update prop limits now to ensure limits exist before
        # widgets are created
        self._set_dim_slider_limits()
        self._set_dim_diff_enabled()

        # Add listeners to the properties which will cause a
        # refresh of the Info Panel and slider limits
        for dim in range(5, 8):
            prop = ds.getProp(f'dim_{dim}')
            prop.addListener(
                ds,
                f'dim_{dim}_info_update',
                self._selectedIndexChanged,
                overwrite=True)

            prop = ds.getProp(f'dim_{dim}_avg')
            prop.addListener(
                ds,
                f'dim_{dim}_avg_slider_update',
                self._set_dim_slider_limits,
                overwrite=True)

            prop.addListener(
                ds,
                f'dim_{dim}_avg_info_update',
                self._selectedIndexChanged,
                overwrite=True)

            prop = ds.getProp(f'dim_{dim}_diff')
            prop.addListener(
                ds,
                f'dim_{dim}_diff_slider_update',
                self._set_dim_slider_limits,
                overwrite=True)

        widgetList = self.getWidgetList()

        widgetList.AddGroup(
            'niftiMRSDimensions',
            'NIfTI-MRS Dimensions ')

        dsWidgets = self.generateDataSeriesWidgets(
            ds,
            'niftiMRSDimensions',
            overlay.ndim)

        self.__dsWidgets = dsWidgets

    def refreshInfoPanel(self):
        '''Create / re-create the NIfTI-MRS information panel'''
        widgetList = self.getWidgetList()

        if widgetList.HasGroup('niftiMRSInfo'):
            widgetList.RemoveGroup('niftiMRSInfo')

        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            return

        ds = self.__plotPanel.getDataSeries(overlay)

        widgetList.AddGroup(
            'niftiMRSInfo',
            'NIfTI-MRS Information ')

        curr_hdr_exts = overlay.header.extensions
        hdr_ext_codes = curr_hdr_exts.get_codes()
        hdr_ext = json.loads(curr_hdr_exts[hdr_ext_codes.index(44)]
                             .get_content())

        def my_static_txt(text):
            st = wx.StaticText(widgetList,
                               label=text,
                               style=wx.ALIGN_LEFT)
            return st

        for dim in range(5, 8):
            if dim <= overlay.ndim:
                tag = hdr_ext[f'dim_{dim}']
                widgetList.AddWidget(
                    my_static_txt(tag),
                    f'dim_{dim} tag',
                    groupName='niftiMRSInfo')

                dim_size = overlay.shape[dim-1]
                widgetList.AddWidget(
                    my_static_txt(f'{dim_size}'),
                    f'dim_{dim} size',
                    groupName='niftiMRSInfo')

                # Optional dim headers
                if f'dim_{dim}_info' in hdr_ext:
                    widgetList.AddWidget(
                        my_static_txt(hdr_ext[f'dim_{dim}_info']),
                        f'dim_{dim}_info',
                        groupName='niftiMRSInfo')

                # Process dynamic header fields
                def interpret_dyn_header(obj, index):
                    if isinstance(obj, dict)\
                            and "Value" in obj:
                        return interpret_dyn_header(obj["Value"], index)
                    elif isinstance(obj, list):
                        return str(obj[index])
                    elif isinstance(obj, dict)\
                            and "start" in obj\
                            and "increment" in obj:
                        return str(obj['start']
                                   + index * obj['increment'])
                    else:
                        raise TypeError('Incorrect type for dynamic header.')

                d_hdr_str = f'dim_{dim}_header'
                if d_hdr_str in hdr_ext:
                    index = getattr(ds, f'dim_{dim}')
                    for key in hdr_ext[d_hdr_str]:
                        dim_hdr_value = interpret_dyn_header(
                            hdr_ext[d_hdr_str][key], index)

                        widgetList.AddWidget(
                            my_static_txt(dim_hdr_value),
                            f'dim_{dim}: {key}',
                            groupName='niftiMRSInfo')

        # Extract Nucleus and SpectrometerFrequency
        nucleus = hdr_ext['ResonantNucleus'][0]
        spec_freq = hdr_ext['SpectrometerFrequency'][0]
        specwidth = 1 / overlay.header['pixdim'][4]

        widgetList.AddWidget(
            my_static_txt(nucleus),
            'Nucleus ',
            groupName='niftiMRSInfo')

        widgetList.AddWidget(
            my_static_txt(f'{spec_freq:0.3f}'),
            'Frequency (MHz) ',
            groupName='niftiMRSInfo')

        widgetList.AddWidget(
            my_static_txt(f'{specwidth:0.0f}'),
            'Spectral width (Hz) ',
            groupName='niftiMRSInfo')

    def _check_nifti_mrs(self):
        '''Check that the overlay selected is a valid NIfTI-MRS overlay'''

        overlay = self.displayCtx.getSelectedOverlay()
        if overlay is None:
            return False

        nifti_mrs_re = re.compile(r'mrs_v\d+_\d+')
        intent_str = overlay.header.get_intent()[2]

        # Check that file is NIfTI-MRS by looking for suitable intent string
        if isinstance(overlay, fslimage.Nifti)\
                and overlay.ndim > 3\
                and nifti_mrs_re.match(intent_str):
            return True
        else:
            return False

    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` or
        :class:`.OverlayList` changes.
        """

        # Double check that the selected overlay has
        # changed before refreshing the panel, as it
        # may not have (e.g. new overlay added, but
        # selected overlay stayed the same).
        if self.displayCtx.getSelectedOverlay() is not self.__selectedOverlay\
                and self._check_nifti_mrs():
            self.refreshDataSeriesWidgets()
            self.refreshInfoPanel()

    def _selectedIndexChanged(self, *a):
        """Called when the dimension indices properties are changed.
        """
        self.refreshInfoPanel()


# Add shortcuts to open MRS controls.
# Currently FSLeyes does not have an
# API for this, so we hack the shortcuts
# into the fsleyes.shortcuts module.
fsleyes_shortcuts.actions['MRSView.MRSToolBar'] = 'Ctrl-Alt-3'
fsleyes_shortcuts.actions['MRSView.MRSControlPanel'] = 'Ctrl-Alt-4'
fsleyes_shortcuts.actions['MRSView.MRSDimControl'] = 'Ctrl-Alt-5'


class MRSView(PowerSpectrumPanel):
    """The ``MRSView`` is a FSLeyes view panel for plotting data from MRS
    NIFTI images.
    """

    @staticmethod
    def title():
        """Overrides :meth:`.ViewPanel.title`. Returns a title to be used
        in FSLeyes menus.
        """
        return 'MRS'

    @staticmethod
    def controlOrder():
        """Overrides :meth:`.ViewPanel.controlOrder`. Returns a suggested
        ordering of control panels for the FSLeyes settings menu.
        """
        return ['OverlayListPanel',
                'PlotListPanel',
                'MRSToolBar',
                'MRSControlPanel',
                'MRSDimControl']

    @staticmethod
    def defaultLayout():
        """Overrides :meth:`.ViewPanel.defaultLayout`. Returns a list
        of control panels that should be added by default for new ``MRSView``
        views.
        """

        return ['OverlayListPanel',
                'PlotListPanel',
                'MRSToolBar',
                'MRSDimControl']

    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``MRSView``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """

        PowerSpectrumPanel.__init__(self,
                                    parent,
                                    overlayList,
                                    displayCtx,
                                    frame)

        self.initProfile(MRSViewProfile)

        self._add_annotation_listener()

    def destroy(self):
        """Must be called when this ``MRSView`` is no longer
        needed. Removes some property listeners, and calls
        :meth:`.OverlayPlotPanel.destroy`.
        """

        self.canvas.removeListener('dataSeries', 'mrs_voxel_annotations')
        PowerSpectrumPanel.destroy(self)

    def draw(self, *a):
        """Overrides :meth:`.PlotPanel.draw`. Draws some
        :class:`.PowerSpectrumSeries` using the
        :meth:`.PlotCanvas.drawDataSeries` method.
        """

        if not self or self.destroyed:
            return

        canvas = self.canvas
        pss = self.getDataSeriesToPlot()

        for ps in pss:
            with props.suppress(ps, 'label'):
                ps.label = ps.makeLabel()

        if len(pss) > 0:
            self._set_mrs_plot_scale(pss)

        canvas.drawDataSeries(extraSeries=pss)
        canvas.drawArtists()

    def createDataSeries(self, overlay):
        """Overrides :meth:`.OverlayPlotPanel.createDataSeries`. Creates a
        :class:`.PowerSpectrumSeries` instance for the given overlay.

        Overload the PowerSpectrumPanel definition to allow only complex
        fslimage.Image and implement the multi-dimensional
        ComplexPowerSpectrumSeries class (MDComplexPowerSpectrumSeries).
        """

        displayCtx = self.displayCtx
        overlayList = self.overlayList

        psargs = [overlay, overlayList, displayCtx, self]

        if isinstance(overlay, fslimage.Image) and overlay.ndim > 3:

            if overlay.iscomplex:
                ps = MDComplexPowerSpectrumSeries(*psargs)
            else:
                return None, None, None

            opts = displayCtx.getOpts(overlay)
            targets = [displayCtx, opts]
            propNames = ['location', 'volumeDim']

        else:
            return None, None, None

        ps.colour = self.getOverlayPlotColour(overlay)
        ps.lineStyle = '-'
        ps.lineWidth = 2
        ps.alpha = 1.0

        return ps, targets, propNames

    def _set_mrs_plot_scale(self, pss):
        # Extract NIfTI-MRS header extensions
        hdr_ext = []
        for ps in pss:
            curr_hdr_exts = ps.overlay.header.extensions
            hdr_ext_codes = curr_hdr_exts.get_codes()
            hdr_ext.append(
                json.loads(
                    curr_hdr_exts[hdr_ext_codes.index(44)].get_content()))

        # Extract Nucleus and SpectrometerFrequency
        nuclei = [he['ResonantNucleus'][0] for he in hdr_ext]
        spec_freq = [he['SpectrometerFrequency'][0] for he in hdr_ext]

        all_ps_match = True
        if len(nuclei) > 1:
            for nn in nuclei[1:]:
                if nn != nuclei[0]:
                    all_ps_match = False
        if len(spec_freq) > 1:
            for sf in spec_freq[1:]:
                if not np.isclose(sf, spec_freq[0], atol=1):
                    all_ps_match = False

        canvas = self.canvas
        if all_ps_match and canvas.xAutoScale:
            canvas.xAutoScale = False

            # Calculate and set x scaling
            canvas.xScale = -1 / spec_freq[0]

            # Apply x offset
            known_nuclei = constants.GYRO_MAG_RATIO.keys()
            if nuclei[0] in known_nuclei:
                canvas.xOffset = constants.PPM_SHIFT[nuclei[0]]
            else:
                print(f'Unknown nucleus {nuclei[0]}.')
                canvas.xOffset = 0.0

            canvas.invertX = True
            if nuclei[0] in known_nuclei:
                extra_range = 0.1 * (constants.PPM_RANGE[nuclei[0]][1] - constants.PPM_RANGE[nuclei[0]][0])
                canvas.limits = [
                    constants.PPM_RANGE[nuclei[0]][1] + extra_range,
                    constants.PPM_RANGE[nuclei[0]][0] - extra_range,
                    canvas.limits[2],
                    canvas.limits[3]]
            else:
                # Force redraw.
                canvas.limits = canvas.limits

    def _add_annotation_listener(self):
        pcanvas = self.canvas

        ortho = self.frame.getView(OrthoPanel)
        if len(ortho) == 0:
            log.error('No Ortho panel present')
            return
        else:
            # Assume only 1 ortho panel
            ortho = ortho[0]

        xcanvas = ortho.getXCanvas()
        ycanvas = ortho.getYCanvas()
        zcanvas = ortho.getZCanvas()
        self.dsannotations = {}

        def dataSeriesColourChanged(*a):
            '''Function called to change an annotation colour if
            the dataseries colour is changed.
            '''
            # Argument 0 is the colour changed to,
            # Argument 2 is the dataseries
            colour = a[0]
            edited_ds = a[2]
            for anno in self.dsannotations[edited_ds]:
                anno.colour = colour

            xcanvas.Refresh()
            ycanvas.Refresh()
            zcanvas.Refresh()

        def dataSeriesChanged(*a):
            '''Function called to add annotations on addition of a dataseries'''
            # remove annotation for any newly removed data series
            for ds in self.dsannotations:
                if ds not in pcanvas.dataSeries:
                    xa, ya, za = self.dsannotations.pop(ds)
                    xcanvas.getAnnotations().dequeue(xa, hold=True)
                    ycanvas.getAnnotations().dequeue(ya, hold=True)
                    zcanvas.getAnnotations().dequeue(za, hold=True)

                    # 2. Remove listener for colour changes
                    ds.removeListener('colour', f'{ds.name}_colour')

                    break  # To avoid changing the dictionary size during iteration.

            # create an annotation for any newly added data series
            for ds in pcanvas.dataSeries:
                if ds not in self.dsannotations:
                    # 1. Add annotation to ortho canvases
                    xpos, ypos, zpos = self.displayCtx.location
                    xa = xcanvas.getAnnotations().ellipse(ypos, zpos, 2, 2, colour=ds.colour, hold=True)
                    ya = ycanvas.getAnnotations().ellipse(xpos, zpos, 2, 2, colour=ds.colour, hold=True)
                    za = zcanvas.getAnnotations().ellipse(xpos, ypos, 2, 2, colour=ds.colour, hold=True)
                    self.dsannotations[ds] = [xa, ya, za]

                    # 2. Add listener for colour changes
                    ds.addListener('colour', f'{ds.name}_colour', dataSeriesColourChanged, weak=False)

            xcanvas.Refresh()
            ycanvas.Refresh()
            zcanvas.Refresh()

        # Add listener for new dataseries
        pcanvas.addListener('dataSeries', 'mrs_voxel_annotations', dataSeriesChanged, weak=False)


class MDComplexPowerSpectrumSeries(psseries.ComplexPowerSpectrumSeries):
    '''Sub class of ComplexPowerSpectrumSeries to overload the
       dataAtCurrentVoxel method'''

    dim_5 = props.Int(default=0, clamped=True)
    dim_5_avg = props.Boolean(default=False)
    dim_5_diff = props.Boolean(default=False)

    dim_6 = props.Int(default=0, clamped=True)
    dim_6_avg = props.Boolean(default=False)
    dim_6_diff = props.Boolean(default=False)

    dim_7 = props.Int(default=0, clamped=True)
    dim_7_avg = props.Boolean(default=False)
    dim_7_diff = props.Boolean(default=False)

    """Higher order dimension indicies. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Needs it's own cache or a hack around the name mangling
        self.__cache = cache.Cache(maxsize=1000)

    # The PlotPanel uses a new thread to access
    # data every time the displaycontext location
    # changes. So we mark this method as mutually
    # exclusive to prevent multiple
    # near-simultaneous accesses to the same voxel
    # location. The first time that a voxel location
    # is accessed, its data is cached. So when
    # subsequent (blocked) accesses execute, they
    # will hit the cache instead of hitting the disk
    # (which is a good thing).
    # @idle.mutex
    def dataAtCurrentVoxel(self):
        """Returns the data for the current voxel of the overlay.  This method
        is intended to be used within the :meth:`DataSeries.getData` method
        of sub-classes.

        An internal cache is used to avoid the need to retrieve data for the
        same voxel multiple times, as retrieving data from large compressed
        4D images can be time consuming.

        The location for the current voxel is calculated by the
        :meth:`currentVoxelLocation` method, and the data lookup is performed
        by the :meth:`currentVoxelData` method. These methods may be
        overridden by sub-classes.

        :returns: A ``numpy`` array containing the data at the current
                  voxel, or ``None`` if the current location is out of bounds
                  of the image.
        """

        location = self.currentVoxelLocation(with_time_dim=False)

        if location is None:
            return None

        try:
            data = self.__cache.get(location, None)
        except TypeError:
            # Handle TypeError: unhashable type: 'slice'
            data = None

        if data is None:
            data = self.currentVoxelData(self.currentVoxelLocation())
            try:
                self.__cache.put(location, data)
            except TypeError:
                # Handle TypeError: unhashable type: 'slice'
                pass

        return data

    def currentVoxelLocation(self, with_time_dim=True):
        """Used by :meth:`dataAtCurrentVoxel`. Returns the current voxel
        location. This is used as a key for the voxel data cache implemented
        within the :meth:`dataAtCurrentVoxel` method, and subsequently passed
        to the :meth:`currentVoxelData` method.

        This implements the higher dimension indexing

        If with_time_dim is True (default) then a slice(None) is inserted as
        a fourth dimension.
        """

        opts = self.displayCtx.getOpts(self.overlay)
        voxel = opts.getVoxel()

        if voxel is None:
            return None

        higher_dim_idx = []
        for idx in range(5, 8):
            if getattr(self, f'dim_{idx}_avg'):
                higher_dim_idx.append(slice(None))
            elif getattr(self, f'dim_{idx}_diff'):
                higher_dim_idx.append(slice(0, 2))
            else:
                higher_dim_idx.append(getattr(self, f'dim_{idx}'))

        if with_time_dim:
            return tuple(voxel + [slice(None), ] + higher_dim_idx)
        else:
            return tuple(voxel + higher_dim_idx)

    def currentVoxelData(self, location):
        """Used by :meth:`dataAtCurrentVoxel`. Returns the data at the
        specified location.

        This method may be overridden by sub-classes.
        """

        data = self.overlay[location].copy()

        # Take mean of averaged dimensions
        reduced_loc = [x for x in location[-3:] if isinstance(x, slice)]
        for idx, loc in enumerate(reduced_loc):
            if loc == slice(None):
                data = np.mean(data, axis=idx+1, keepdims=True)
            elif loc == slice(0, 2):
                data = np.diff(data, axis=idx+1)

        data = calcSpectrum(data.squeeze())

        return data


def calcSpectrum(data):
    """Calculates a spectrum for the given one-dimensional data array.
    Includes scaling of first FID point.

    :arg data:    Numpy array containing the time series data

    :returns:     The complex spectrum is returned.
    """

    # Fourier transform on complex data
    data[0] *= 0.5
    data = fft.fft(data)
    data = fft.fftshift(data)

    return data


###############################
# Define a default mrs layout #
###############################

mrs_fsleyes_layout = """
fsleyes.views.orthopanel.OrthoPanel,fsleyes_plugin_mrs.plugin.MRSView
layout2|name=OrthoPanel 1;caption=Ortho View 1;state=67377088;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=MRSView 2;caption=MRS view 2;state=67377148;dir=2;layer=0;row=1;pos=0;prop=100000;bestw=1296;besth=262;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=22|dock_size(2,0,1)=648|
fsleyes.controls.orthotoolbar.OrthoToolBar,fsleyes.controls.overlaydisplaytoolbar.OverlayDisplayToolBar,fsleyes.controls.overlaylistpanel.OverlayListPanel,fsleyes.controls.locationpanel.LocationPanel;syncLocation=True,syncOverlayOrder=True,syncOverlayDisplay=True,syncOverlayVolume=True,movieRate=400,movieAxis=3;showCursor=True,bgColour=#000000ff,fgColour=#ffffffff,cursorColour=#00ff00ff,cursorGap=False,showColourBar=False,colourBarLocation=top,colourBarLabelSide=top-left,showXCanvas=True,showYCanvas=True,showZCanvas=True,showLabels=True,labelSize=12,layout=grid,xzoom=100.0,yzoom=100.0,zzoom=100.0
layout2|name=Panel;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OrthoToolBar;caption=Ortho view toolbar;state=67382012;dir=1;layer=10;row=0;pos=0;prop=100000;bestw=607;besth=35;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OverlayDisplayToolBar;caption=Display toolbar;state=67382012;dir=1;layer=11;row=0;pos=0;prop=100000;bestw=922;besth=49;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OverlayListPanel;caption=Overlay list;state=67373052;dir=3;layer=0;row=0;pos=0;prop=100000;bestw=201;besth=84;minw=1;minh=1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=201;floath=100;notebookid=-1;transparent=255|name=LocationPanel;caption=Location;state=67373052;dir=3;layer=0;row=0;pos=1;prop=100000;bestw=383;besth=111;minw=1;minh=1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=383;floath=127;notebookid=-1;transparent=255|dock_size(5,0,0)=22|dock_size(3,0,0)=139|dock_size(1,10,0)=37|dock_size(1,11,0)=51|
fsleyes_plugin_mrs.plugin.MRSToolBar,fsleyes_plugin_mrs.plugin.MRSDimControl,fsleyes.controls.overlaylistpanel.OverlayListPanel,fsleyes.controls.plotlistpanel.PlotListPanel;;
layout2|name=FigureCanvasWxAgg;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=640;besth=480;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=MRSToolBar;caption=MRS toolbar;state=67382012;dir=1;layer=10;row=0;pos=0;prop=100000;bestw=272;besth=34;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=MRSDimControl;caption=NIfTI-MRS;state=67373052;dir=2;layer=0;row=0;pos=0;prop=100000;bestw=188;besth=144;minw=1;minh=1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=188;floath=160;notebookid=-1;transparent=255|name=OverlayListPanel;caption=Overlay list;state=67373052;dir=2;layer=0;row=0;pos=1;prop=100000;bestw=201;besth=52;minw=1;minh=1;maxw=-1;maxh=-1;floatx=1194;floaty=492;floatw=201;floath=68;notebookid=-1;transparent=255|name=PlotListPanel;caption=Plot list;state=67373052;dir=2;layer=0;row=0;pos=2;prop=100000;bestw=201;besth=52;minw=1;minh=1;maxw=-1;maxh=-1;floatx=1212;floaty=692;floatw=201;floath=68;notebookid=-1;transparent=255|dock_size(5,0,0)=642|dock_size(1,10,0)=36|dock_size(2,0,0)=260|
""".strip()  # noqa: E501


# FSLeyes >= 1.8 allows us to specify layouts as entry points (see setup.py).
# In older versions of FSLeyes, we have to abuse the BUILT_IN_LAYOUTS
# to embed an MRS view by default
def parseVersion(version):
    parts = []
    for part in version.split('.'):
        # Give up if we can't parse a component - in case
        # we have a development version, e.g. "2.0.0.dev0"
        try:
            parts.append(int(part))
        except:  # noqa: E722
            break
    return tuple(parts)


if parseVersion(fsleyes_version) < (1, 8, 0):
    BUILT_IN_LAYOUTS.update({'mrs': mrs_fsleyes_layout}) # noqa



"""Tests for basic plugin functionality.
Here we test that fsleyes loads without error when the plugin is present.
We also test that it loads with the default mrs layout.
"""

from pathlib import Path
from subprocess import check_call

# Data paths
test_data_path = Path(__file__).parent / 'test_data'
t1 = test_data_path / 'T1.nii.gz'
svs = test_data_path / 'metab.nii.gz'


def test_load(tmp_path):
    assert not check_call(['render',
                           '-of', tmp_path / 'test.png',
                           str(t1),
                           str(svs)])


def test_load_mrs_layout(tmp_path):
    assert not check_call(['render',
                           '-of', tmp_path / 'test.png',
                           '-smrs',
                           str(t1),
                           str(svs)])



#!/usr/bin/env python

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='fsleyes-plugin-mrs',

    version='0.1.2',

    description='FSLeyes extension for viewing MRS(I) data formatted as NIfTI-MRS.',
    author='William Clarke, University of Oxford',
    author_email='william.clarke@ndcn.ox.ac.uk',
    url='https://git.fmrib.ox.ac.uk/wclarke/fsleyes-plugin-mrs',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent"],

    python_requires='>=3.7',

    packages=['fsleyes_plugin_mrs'],

    entry_points={

        'fsleyes_views': [
            'MRS view = fsleyes_plugin_mrs.plugin:MRSView',
        ],

        'fsleyes_controls': [
            'NIfTI-MRS = fsleyes_plugin_mrs.plugin:MRSDimControl',
            'MRS control = fsleyes_plugin_mrs.plugin:MRSControlPanel',
            'MRS toolbar = fsleyes_plugin_mrs.plugin:MRSToolBar',
            'FSL-MRS Results = fsleyes_plugin_mrs.results_load:FSLMRSResultsControl',
        ],

        'fsleyes_tools': [
            'Load FSL-MRS fit = fsleyes_plugin_mrs.results_load:FSLFitTool',
        ],

        'fsleyes_layouts': [
            'mrs = fsleyes_plugin_mrs.plugin:mrs_fsleyes_layout'
        ]
    },

    package_data={'fsleyes_plugin_mrs': ['icons/*.png']},

    install_requires=['fsleyes>=1.0.10']
)



#!/usr/bin/env python
#
# mrsviewprofile.py - The MRSViewProfile class.
#
# Author: Will Clarke <william.clarke@ndcn.ox.ac.uk>
#
"""This module provides the :class:`.MRSViewProfile` class, an interaction
profile for the :class:`.MRSView`.
"""

import wx

import fsl.data.image as fslimage
from fsleyes.profiles import plotprofile

from matplotlib.text import Text


class MRSViewProfile(plotprofile.PlotProfile):
    """The ``MRSViewProfile`` is a :class:`.PlotProfile` for use with the
    :class:`.MRSView`.

    In addition to the ``panzoom`` mode provided by the :class:`.PlotProfile`
    class, the ``MRSViewProfile`` class implements a ``phasing`` mode, in
    which the user is able to click/drag on a plot to change the
    zero and first order phase for the currently selected overlay.


    """

    @staticmethod
    def supportedView():
        """Returns the :class:`.MRSView` class. """
        from .plugin import MRSView
        return MRSView

    @staticmethod
    def tempModes():
        """Returns the temporary mode map for the ``MRSViewProfile``,
        which controls the use of modifier keys to temporarily enter other
        interaction modes.
        """
        return {('panzoom', wx.WXK_CONTROL): 'phasing0',
                ('panzoom', wx.WXK_SHIFT): 'phasing1',
                ('panzoom', (wx.WXK_CONTROL, wx.WXK_SHIFT)): 'phasingBoth',
                ('panzoom', (wx.WXK_ALT, wx.WXK_CONTROL)): 'phasing0_all',
                ('panzoom', (wx.WXK_ALT, wx.WXK_SHIFT)): 'phasing1_all',
                ('panzoom', (wx.WXK_ALT, wx.WXK_CONTROL, wx.WXK_SHIFT)): 'phasingBoth_all'}

    def __init__(self, viewPanel, overlayList, displayCtx):
        """Create a ``MRSViewProfile``.

        :arg viewPanel:    A :class:`.MRSView` instance.

        :arg overlayList:  The :class:`.OverlayList` instance.

        :arg displayCtx:   The :class:`.DisplayContext` instance.
        """
        plotprofile.PlotProfile.__init__(self,
                                         viewPanel,
                                         overlayList,
                                         displayCtx,
                                         ['phasing0', 'phasing1', 'phasing'])

        self._guideText = None
        self._p0 = 0
        self._p1 = 0
        self._tempmodetext = None

    def __phasingModeCompatible(self):
        """Returns ``True`` if phasing can currently be carried out, ``False``
        otherwise.
        """

        overlay = self.displayCtx.getSelectedOverlay()

        if not isinstance(overlay, fslimage.Image):
            return False

        if not overlay.iscomplex:
            return False

        if len(overlay.shape) < 4 or overlay.shape[3] == 1:
            return False

        return True

    def __updatePhase(self, xvalue, yvalue, p0, p1, all):
        """Called by the ``phasing`` event handlers.
        Updates the zeroth and first order phase.
        :arg xvalue: Normalised x position (0 to 1)
        :arg yvalue: Normalised y position (0 to 1)
        :arg bool p0: If true then 0th order phase will be updated
        :arg bool p1: If true then 1st order phase will be updated
        :arg bool all: If true then all dataseries will be updated, else only the selected.
        """
        if xvalue is None or yvalue is None:
            return

        if p0:
            self._p0 = 1.1 * 360 * (xvalue - 0.5)
        if p1:
            self._p1 = 0.004 * (yvalue - 0.5)

        mrsPanel = self.viewPanel
        if all:
            # Loop through all dataseries to update the phase
            for ds in mrsPanel.getDataSeriesToPlot():
                ds.zeroOrderPhaseCorrection = self._p0
                ds.firstOrderPhaseCorrection = self._p1
        else:
            # Identify the single spectrum to update.
            overlay = self.displayCtx.getSelectedOverlay()
            spec = mrsPanel.getDataSeries(overlay)

            # Update
            spec.zeroOrderPhaseCorrection = self._p0
            spec.firstOrderPhaseCorrection = self._p1

    def __createTextString(self):
        '''Create the text for the on-screen guide'''
        p1_ms = 1000 * self._p1
        return f'Mode: {self._tempmodetext}\n'\
               f'0th: {self._p0:0.1f} degrees, 1st: {p1_ms:0.2f} ms\n'\
               'Use ctrl for 0th only (left/right).\n'\
               'Use shift for 1st only (up/down).\n'\
               'Use ctrl+shift for both (l/r/u/d).\n'\
               '+ alt to phase all.'

    def __createText(self, draw=True):
        """Create the on screen text guide.
        :arg bool draw: If true then draw will be called at end. Defaults to True.
        """
        text = self.__createTextString()
        # Create text, use pixel positioning.
        self._guideText = Text(20, 20, text, transform=None)

        # Add to artists and draw
        canvas = self.viewPanel.canvas
        canvas.artists.append(self._guideText)
        if draw:
            canvas.drawArtists(immediate=True)

    def __removeText(self, draw=True):
        """Remove the on screen text guide.
        :arg bool draw: If true then draw will be called at end. Defaults to True.
        """
        mpl_canvas = self.viewPanel.canvas
        if self._guideText in mpl_canvas.artists:
            mpl_canvas.artists.remove(self._guideText)
            if draw:
                mpl_canvas.drawArtists(immediate=True)

    def __updateText(self):
        """Update (remove and recreate) the on screen text guide.
        Draw not called, wait for an update.
        """
        self.__removeText(draw=False)
        self.__createText(draw=False)

    def _phasingModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """On first mouse down, draws on screen text guide. """

        if not self.__phasingModeCompatible():
            return

        # Create text for first time
        self.__createText()

    def _phasingModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos, p0=True, p1=True, all=False):
        """Updates the phase of the selected overlay and on screen text guide. """
        if mousePos is None:
            xvalue, yvalue = None, None
        else:
            canvas_x, canvas_y = canvas.get_width_height()
            xvalue = mousePos[0] / canvas_x
            yvalue = mousePos[1] / canvas_y

        self.__updatePhase(xvalue, yvalue, p0, p1, all)

        self.__updateText()

    def _phasingModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """On final mouse up, removes on screen text guide. """
        # Clear the temp mode text.
        self._tempmodetext = None

        # Remove text at end of phasing
        self.__removeText()

    # Handle all the temp mode options
    # 0th order only - phasing0
    def _phasing0ModeLeftMouseDown(self, *a, **kwa):
        self._tempmodetext = '0th order only'
        self._phasingModeLeftMouseDown(*a, **kwa)

    def _phasing0ModeLeftMouseDrag(self, *a, **kwa):
        self._phasingModeLeftMouseDrag(*a, p0=True, p1=False, **kwa)

    def _phasing0ModeLeftMouseUp(self, *a, **kwa):
        self._phasingModeLeftMouseUp(*a, **kwa)

    # 1st order only - phasing1
    def _phasing1ModeLeftMouseDown(self, *a, **kwa):
        self._tempmodetext = '1st order only'
        self._phasingModeLeftMouseDown(*a, **kwa)

    def _phasing1ModeLeftMouseDrag(self, *a, **kwa):
        self._phasingModeLeftMouseDrag(*a, p0=False, p1=True, **kwa)

    def _phasing1ModeLeftMouseUp(self, *a, **kwa):
        self._phasingModeLeftMouseUp(*a, **kwa)

    # 0th and 1st order - phasingBoth
    def _phasingBothModeLeftMouseDown(self, *a, **kwa):
        self._tempmodetext = '0th & 1st order'
        self._phasingModeLeftMouseDown(*a, **kwa)

    def _phasingBothModeLeftMouseDrag(self, *a, **kwa):
        self._phasingModeLeftMouseDrag(*a, **kwa)

    def _phasingBothModeLeftMouseUp(self, *a, **kwa):
        self._phasingModeLeftMouseUp(*a, **kwa)

    # 0th order only, all dataseries - phasing0_all
    def _phasing0_allModeLeftMouseDown(self, *a, **kwa):
        self._tempmodetext = '0th order only, all'
        self._phasingModeLeftMouseDown(*a, **kwa)

    def _phasing0_allModeLeftMouseDrag(self, *a, **kwa):
        self._phasingModeLeftMouseDrag(*a, p0=True, p1=False, all=True, **kwa)

    def _phasing0_allModeLeftMouseUp(self, *a, **kwa):
        self._phasingModeLeftMouseUp(*a, **kwa)

    # 1st order only, all dataseries - phasing1_all
    def _phasing1_allModeLeftMouseDown(self, *a, **kwa):
        self._tempmodetext = '1st order only, all'
        self._phasingModeLeftMouseDown(*a, **kwa)

    def _phasing1_allModeLeftMouseDrag(self, *a, **kwa):
        self._phasingModeLeftMouseDrag(*a, p0=False, p1=True, all=True, **kwa)

    def _phasing1_allModeLeftMouseUp(self, *a, **kwa):
        self._phasingModeLeftMouseUp(*a, **kwa)

    # 0th and 1st order, all dataseries - phasingBoth_all
    def _phasingBoth_allModeLeftMouseDown(self, *a, **kwa):
        self._tempmodetext = '0th & 1st order, all'
        self._phasingModeLeftMouseDown(*a, **kwa)

    def _phasingBoth_allModeLeftMouseDrag(self, *a, **kwa):
        self._phasingModeLeftMouseDrag(*a, all=True, **kwa)

    def _phasingBoth_allModeLeftMouseUp(self, *a, **kwa):
        self._phasingModeLeftMouseUp(*a, **kwa)
